@@ -4,7 +4,7 @@ import shutil
 
 import tomli
 
-from julienne.parser import Line, range_token, chapter_in_range
+from julienne.parser import parse_content, range_token, chapter_in_range
 
 # ===========================================================================
 
@@ -35,7 +35,7 @@ class ConditionalDirNode(DirNode):
         self.lower, self.upper = range_token(token)
 
     def info(self):
-        print(f'ConditionalDirNode {self.lower}-{self.upper}')
+        print(f'ConditionalDirNode {self.lower} - {self.upper}')
         print(f'   {self.path}')
 
     def should_traverse(self, chapter):
@@ -62,7 +62,7 @@ class CopyOnlyFileNode:
         rel = self.path.relative_to(base_path)
         dest = output_path / rel
 
-        shutil.copy(self.path, dest)
+        shutil.copy2(self.path, dest)
 
 
 class FileNode:
@@ -81,39 +81,17 @@ class FileNode:
 
         :param content: string to parse
         """
-        if content and content[-1] == '\n':
-            content = content[:-1]
-
-        self.lines = []
-        self.all_conditional = True
-        block_header = None
-        boundary_set = False
-        for item in content.split('\n'):
-            line = Line(item, block_header)
-
-            block_header = line.block_header
-            self.lines.append(line)
-
-            if not line.conditional:
-                self.all_conditional = False
-
-            if boundary_set and line.conditional:
-                # Check if this line changes the boundary conditions
-                if line.lower < self.lowest:
-                    self.lowest = line.lower
-
-                if line.upper == -1 or \
-                        (line.upper != -1 and line.upper > self.highest):
-                    self.highest = line.upper
-            else:
-                # Boundary not set yet
-                if line.conditional:
-                    boundary_set = True
-                    self.lowest = line.lower
-                    self.highest = line.upper
+        self.parser = parse_content(content)
+        self.lowest = self.parser.lowest
+        self.highest = self.parser.highest
+        self.all_conditional = self.parser.all_conditional
 
     def info(self):
-        print('FileNode', self.lowest, self.highest, self.all_conditional)
+        lowest = getattr(self, 'lowest', "Unset")
+        highest = getattr(self, 'highest', "Unset")
+        all_cond = getattr(self, 'all_conditional', "Unset")
+
+        print('FileNode', lowest, highest, all_cond)
         print(f'   {self.path}')
 
     def copy(self, chapter, base_path, output_path):
@@ -127,31 +105,48 @@ class FileNode:
             # Write file if within chapter range
             dest = output_path / rel
             with open(dest, "w") as f:
-                for line in self.lines:
+                for line in self.parser.lines:
                     content = line.get_content(chapter)
                     if content is not None:
                         f.write(content + "\n")
 
+
+class ConditionalFileNode(FileNode):
+    def __init__(self, path, token):
+        self.lower, self.upper = range_token(token)
+        super().__init__(path)
+
+    def info(self):
+        print(f'ConditionalFileNode {self.lower} - {self.upper}')
+        print(f'   {self.path}')
+
+    def copy(self, chapter, base_path, output_path):
+        if not chapter_in_range(chapter, True, self.lower, self.upper):
+            return
+
+        super().copy(chapter, base_path, output_path)
+
 # ---------------------------------------------------------------------------
 
 class NodeFilter:
     def __init__(self):
         self.python_files = []
-        self.conditional_dirs = []
-        self.conditional_map = {}
         self.ignore_dirs = []
+        self.ignore_substrings = []
+
+        self.ranged_files_map = {}
 
     def set_python_filter(self, py_globs, base_dir):
         for py_glob in py_globs:
             self.python_files.extend(base_dir.glob(py_glob))
 
-    def set_dir_filter(self, subdir, base_path):
-        for dir_spec in subdir.values():
-            token = dir_spec['range']
-            dir_path = _convert_path(base_path, Path(dir_spec['src_dir']))
-
-            self.conditional_dirs.append(dir_path)
-            self.conditional_map[dir_path] = token
+    def set_ranged_file_filter(self, ranged_files, base_path):
+        # Loop over all the "ranged_files" maps and track their content
+        for spec in ranged_files.values():
+            token = spec['range']
+            for path in spec['files']:
+                ranged_path = _convert_path(base_path, Path(path))
+                self.ranged_files_map[ranged_path] = token
 
     def set_ignore_dirs(self, ignore_dirs, base_path):
         for dirname in ignore_dirs:
@@ -164,12 +159,22 @@ class NodeFilter:
 
 def _process_directory(parent, dir_path, node_filter):
     for path in dir_path.iterdir():
+        # Skip any paths that are in our ignore_substrings list
+        skip = False
+        for ignore_substring in node_filter.ignore_substrings:
+            if ignore_substring in str(path):
+                skip = True
+
+        if skip:
+            # Path contained something spec'd in ignore_substrings, ignore it
+            continue
+
         try:
             if path.is_dir():
                 if path in node_filter.ignore_dirs:
                     continue
-                elif path in node_filter.conditional_dirs:
-                    token = node_filter.conditional_map[path]
+                elif path in node_filter.ranged_files_map.keys():
+                    token = node_filter.ranged_files_map[path]
                     node = ConditionalDirNode(path, token)
                 else:
                     node = DirNode(path)
@@ -178,7 +183,12 @@ def _process_directory(parent, dir_path, node_filter):
                 _process_directory(node, node.path, node_filter)
             else:
                 if path in node_filter.python_files:
-                    node = FileNode(path)
+                    if path in node_filter.ranged_files_map.keys():
+                        token = node_filter.ranged_files_map[path]
+                        node = ConditionalFileNode(path, token)
+                    else:
+                        node = FileNode(path)
+
                     node.parse_file()
                 else:
                     node = CopyOnlyFileNode(path)
@@ -202,11 +212,15 @@ def _traverse(chapter, node, cmd, *args):
 
 def _find_biggest(node, biggest=1):
     result = biggest
+    if hasattr(node, "lower") and node.lower > result:
+        result = node.lower
+    if hasattr(node, "upper") and node.upper != -1 and node.upper > result:
+        result = node.upper
 
     try:
         for child in node.children:
             if isinstance(child, DirNode):
-                subresult = _find_biggest(child, biggest)
+                subresult = _find_biggest(child, result)
                 if subresult > result:
                     result = subresult
             elif isinstance(child, FileNode):
@@ -259,18 +273,21 @@ def generate_files(config_file):
     py_globs = config.get('python_globs', ['**/*.py', ])
     node_filter.set_python_filter(py_globs, base_dir)
 
-    subdir = config.get('subdir', {})
-    node_filter.set_dir_filter(subdir, base_path)
+    ranged_files = config.get('ranged_files', {})
+    node_filter.set_ranged_file_filter(ranged_files, base_path)
 
     ignore_dirs = config.get('ignore_dirs', [])
     node_filter.set_ignore_dirs(ignore_dirs, base_dir)
 
+    node_filter.ignore_substrings = config.get('ignore_substrings', [])
+
     # Create node structure
     root = DirNode(base_dir)
     _process_directory(root, base_dir, node_filter)
 
     # DEBUG:
     #_traverse(3, root, 'info')
+    #exit()
 
     ### Generate output
     parent_path = base_dir.parent
@@ -284,6 +301,7 @@ def generate_files(config_file):
     #num = 5
     #output_path = output_dir / Path(f"ch{num}")
     #_traverse(num, root, 'copy', num, parent_path, output_path)
+    #exit()
 
     for num in range(1, biggest + 1):
         print(f'Creating chapter {num}')
